#!/usr/bin/env python3
import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__) + "/.."))

import asyncio
import json
from argparse import OPTIONAL, ArgumentParser

from dbus_fast import BusType, Message, MessageType, Variant
from dbus_fast.aio import MessageBus
from dbus_fast.validators import (
    is_bus_name_valid,
    is_interface_name_valid,
    is_member_name_valid,
    is_object_path_valid,
)

parser = ArgumentParser()

parser.add_argument("--system", help="Use the system bus", action="store_true")
parser.add_argument("--session", help="Use the session bus", action="store_true")
parser.add_argument(
    "--dest", help="The destination address for the message", required=True
)
parser.add_argument("--signature", help="The signature for the message body")
parser.add_argument(
    "--type",
    help="The type of message to send",
    choices=[e.name for e in MessageType],
    default=MessageType.METHOD_CALL.name,
    nargs=OPTIONAL,
)
parser.add_argument("object_path", help="The object path for the message")
parser.add_argument("interface.member", help="The interface and member for the message")
parser.add_argument(
    "body",
    help="The JSON encoded body of the message. Must match the signature",
    nargs=OPTIONAL,
)

args = parser.parse_args()


def exit_error(message):
    parser.print_usage()
    print()
    print(message)
    sys.exit(1)


interface_member = vars(args)["interface.member"].split(".")

if len(interface_member) < 2:
    exit_error(
        f"Expecting an interface and member separated by a dot: {vars(args)['interface.member']}"
    )

destination = args.dest
member = interface_member[-1]
interface = ".".join(interface_member[: len(interface_member) - 1])
object_path = args.object_path
signature = args.signature
body = args.body
message_type = MessageType[args.type]
signature = args.signature

bus_type = BusType.SESSION

if args.system:
    bus_type = BusType.SYSTEM

if message_type is not MessageType.METHOD_CALL:
    exit_error("only message type METHOD_CALL is supported right now")

if not is_bus_name_valid(destination):
    exit_error(f"got invalid bus name: {destination}")

if not is_object_path_valid(object_path):
    exit_error(f"got invalid object path: {object_path}")

if not is_interface_name_valid(interface):
    exit_error(f"got invalid interface name: {interface}")

if not is_member_name_valid(member):
    exit_error(f"got invalid member name: {member}")

if body is None:
    body = []
    signature = ""
else:
    try:
        body = json.loads(body)
    except json.JSONDecodeError as e:
        exit_error(f"could not parse body as JSON: ({e})")

    if type(body) is not list:
        exit_error("body must be an array of arguments")

    if not signature:
        exit_error("--signature is a required argument when passing a message body")


async def main():
    bus = await MessageBus(bus_type=bus_type).connect()

    message = Message(
        destination=destination,
        member=member,
        interface=interface,
        path=object_path,
        signature=signature,
        body=body,
    )

    result = await bus.call(message)

    ret = 0

    if result.message_type is MessageType.ERROR:
        print(f"Error: {result.error_name}", file=sys.stderr)
        ret = 1

    def default(o):
        if type(o) is Variant:
            return [o.signature, o.value]
        raise json.JSONDecodeError()

    print(json.dumps(result.body, indent=2, default=default))

    sys.exit(ret)


asyncio.run(main())



import pytest

from dbus_fast import (
    is_bus_name_valid,
    is_interface_name_valid,
    is_member_name_valid,
    is_object_path_valid,
)
from dbus_fast.errors import (
    InvalidBusNameError,
    InvalidInterfaceNameError,
    InvalidMemberNameError,
    InvalidObjectPathError,
)
from dbus_fast.validators import (
    assert_bus_name_valid,
    assert_interface_name_valid,
    assert_member_name_valid,
    assert_object_path_valid,
)


def test_object_path_validator():
    valid_paths = ["/", "/foo", "/foo/bar", "/foo/bar/bat"]
    invalid_paths = [
        None,
        "",
        "foo",
        "foo/bar",
        "/foo/bar/",
        "/$/foo/bar",
        "/foo//bar",
        "/foo$bar/baz",
    ]

    for path in valid_paths:
        assert is_object_path_valid(path), f'path should be valid: "{path}"'
    for path in invalid_paths:
        assert not is_object_path_valid(path), f'path should be invalid: "{path}"'


def test_bus_name_validator():
    valid_names = [
        "foo.bar",
        "foo.bar.bat",
        "_foo._bar",
        "foo.bar69",
        "foo.bar-69",
        "org.mpris.MediaPlayer2.google-play-desktop-player",
    ]
    invalid_names = [
        None,
        "",
        "5foo.bar",
        "foo.6bar",
        ".foo.bar",
        "bar..baz",
        "$foo.bar",
        "foo$.ba$r",
    ]

    for name in valid_names:
        assert is_bus_name_valid(name), f'bus name should be valid: "{name}"'
    for name in invalid_names:
        assert not is_bus_name_valid(name), f'bus name should be invalid: "{name}"'


def test_interface_name_validator():
    valid_names = ["foo.bar", "foo.bar.bat", "_foo._bar", "foo.bar69"]
    invalid_names = [
        None,
        "",
        "5foo.bar",
        "foo.6bar",
        ".foo.bar",
        "bar..baz",
        "$foo.bar",
        "foo$.ba$r",
        "org.mpris.MediaPlayer2.google-play-desktop-player",
    ]

    for name in valid_names:
        assert is_interface_name_valid(name), (
            f'interface name should be valid: "{name}"'
        )
    for name in invalid_names:
        assert not is_interface_name_valid(name), (
            f'interface name should be invalid: "{name}"'
        )


def test_member_name_validator():
    valid_members = ["foo", "FooBar", "Bat_Baz69", "foo-bar"]
    invalid_members = [None, "", "foo.bar", "5foo", "foo$bar"]

    for member in valid_members:
        assert is_member_name_valid(member), f'member name should be valid: "{member}"'
    for member in invalid_members:
        assert not is_member_name_valid(member), (
            f'member name should be invalid: "{member}"'
        )


def test_bus_name_requires_dot() -> None:
    """A single-element well-known bus name (no '.') is invalid."""
    assert not is_bus_name_valid("foo")


def test_unique_bus_name_is_valid() -> None:
    """Unique bus names start with ':' and bypass element validation."""
    assert is_bus_name_valid(":1.42")
    assert is_bus_name_valid(":1.0")


def test_bus_name_max_length() -> None:
    """Bus names must not exceed 255 characters."""
    name_255 = "a." + "b" * 253
    name_256 = "a." + "b" * 254
    assert len(name_255) == 255
    assert len(name_256) == 256
    assert is_bus_name_valid(name_255)
    assert not is_bus_name_valid(name_256)


def test_interface_name_max_length() -> None:
    """Interface names must not exceed 255 characters."""
    name_255 = "a." + "b" * 253
    name_256 = "a." + "b" * 254
    assert is_interface_name_valid(name_255)
    assert not is_interface_name_valid(name_256)


def test_interface_name_requires_dot() -> None:
    """A single-element name without '.' is not a valid interface name."""
    assert not is_interface_name_valid("foo")


def test_member_name_max_length() -> None:
    """Member names must not exceed 255 characters."""
    member_255 = "a" * 255
    member_256 = "a" * 256
    assert is_member_name_valid(member_255)
    assert not is_member_name_valid(member_256)


def test_assert_bus_name_valid() -> None:
    """assert_bus_name_valid returns None for valid names, raises for invalid."""
    assert assert_bus_name_valid("foo.bar") is None
    assert assert_bus_name_valid(":1.42") is None
    with pytest.raises(InvalidBusNameError):
        assert_bus_name_valid("5foo.bar")
    with pytest.raises(InvalidBusNameError):
        assert_bus_name_valid(".foo")


def test_assert_object_path_valid() -> None:
    """assert_object_path_valid returns None for valid paths, raises for invalid."""
    assert assert_object_path_valid("/") is None
    assert assert_object_path_valid("/foo/bar") is None
    with pytest.raises(InvalidObjectPathError):
        assert_object_path_valid("foo")
    with pytest.raises(InvalidObjectPathError):
        assert_object_path_valid("/foo/bar/")


def test_assert_interface_name_valid() -> None:
    """assert_interface_name_valid returns None for valid names, raises for invalid."""
    assert assert_interface_name_valid("foo.bar") is None
    with pytest.raises(InvalidInterfaceNameError):
        assert_interface_name_valid("foo")
    with pytest.raises(InvalidInterfaceNameError):
        assert_interface_name_valid("5foo.bar")


def test_assert_member_name_valid() -> None:
    """assert_member_name_valid returns None for valid names, raises for invalid."""
    assert assert_member_name_valid("Foo") is None
    assert assert_member_name_valid("foo-bar") is None
    with pytest.raises(InvalidMemberNameError):
        assert_member_name_valid("5foo")
    with pytest.raises(InvalidMemberNameError):
        assert_member_name_valid("foo.bar")



from __future__ import annotations

import array
import asyncio
import contextlib
import inspect
import logging
import socket
from collections import deque
from collections.abc import Callable
from copy import copy
from functools import partial
from typing import Any
from warnings import warn

from .. import introspection as intr
from ..auth import (
    _AUTH_READ_CHUNK,
    _MAX_AUTH_LINE,
    Authenticator,
    AuthExternal,
)
from ..constants import (
    BusType,
    MessageFlag,
    MessageType,
    NameFlag,
    ReleaseNameReply,
    RequestNameReply,
)
from ..errors import AuthError
from ..message import Message
from ..message_bus import BaseMessageBus, _block_unexpected_reply
from ..service import ServiceInterface, _Method
from .message_reader import build_message_reader
from .proxy_object import ProxyObject

NO_REPLY_EXPECTED_VALUE = MessageFlag.NO_REPLY_EXPECTED.value

_LOGGER = logging.getLogger(__name__)


def _generate_hello_serialized(next_serial: int) -> bytes:
    return bytes(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="Hello",
            serial=next_serial,
        )._marshall(False)
    )


HELLO_1_SERIALIZED = _generate_hello_serialized(1)


def _future_set_exception(fut: asyncio.Future, exc: Exception) -> None:
    if fut is not None and not fut.done():
        fut.set_exception(exc)


def _future_set_result(fut: asyncio.Future, result: Any) -> None:
    if fut is not None and not fut.done():
        fut.set_result(result)


class _MessageWriter:
    """A class to handle writing messages to the message bus."""

    def __init__(self, bus: MessageBus) -> None:
        """A class to handle writing messages to the message bus."""
        self.messages: deque[
            tuple[bytearray, list[int] | None, asyncio.Future | None]
        ] = deque()
        self.negotiate_unix_fd = bus._negotiate_unix_fd
        self.bus = bus
        self.sock = bus._sock
        self.loop = bus._loop
        self.buf: memoryview | None = None
        self.fd = bus._fd
        self.offset = 0
        self.unix_fds: list[int] | None = None
        self.fut: asyncio.Future | None = None

    def write_callback(self, remove_writer: bool = True) -> None:
        """The callback to write messages to the message bus."""
        sock = self.sock
        try:
            while True:
                if self.buf is None:
                    # If there is no buffer, get the next message
                    if not self.messages:
                        # nothing more to write
                        if remove_writer:
                            self.loop.remove_writer(self.fd)
                        return

                    # Get the next message
                    buf, unix_fds, fut = self.messages.popleft()
                    self.unix_fds = unix_fds
                    self.buf = memoryview(buf)
                    self.offset = 0
                    self.fut = fut

                if self.unix_fds and self.negotiate_unix_fd:
                    ancdata = [
                        (
                            socket.SOL_SOCKET,
                            socket.SCM_RIGHTS,
                            array.array("i", self.unix_fds),
                        )
                    ]
                    self.offset += sock.sendmsg([self.buf[self.offset :]], ancdata)
                    self.unix_fds = None
                else:
                    self.offset += sock.send(self.buf[self.offset :])

                if self.offset < len(self.buf):
                    # wait for writable
                    return

                # finished writing
                self.buf = None
                _future_set_result(self.fut, None)
        except Exception as e:
            if self.bus._user_disconnect:
                _future_set_result(self.fut, None)
            else:
                _future_set_exception(self.fut, e)
            self.bus._finalize(e)

    def buffer_message(
        self, msg: Message, future: asyncio.Future | None = None
    ) -> None:
        """Buffer a message to be sent later."""
        unix_fds = msg.unix_fds
        self.messages.append(
            (
                msg._marshall(self.negotiate_unix_fd),
                copy(unix_fds) if unix_fds else None,
                future,
            )
        )

    def _write_without_remove_writer(self) -> None:
        """Call the write callback without removing the writer."""
        self.write_callback(remove_writer=False)

    def schedule_write(
        self, msg: Message | None = None, future: asyncio.Future | None = None
    ) -> None:
        """Schedule a message to be written."""
        queue_is_empty = not self.messages
        if msg is not None:
            self.buffer_message(msg, future)

        if self.bus.unique_name:
            # Optimization: try to send now if the queue
            # is empty. With bleak this usually means we
            # can send right away 99% of the time which
            # is a huge improvement in latency.
            if queue_is_empty:
                self._write_without_remove_writer()

            if (
                self.buf is not None
                or self.messages
                or not self.fut
                or not self.fut.done()
            ):
                self.loop.add_writer(self.fd, self.write_callback)


class MessageBus(BaseMessageBus):
    """The message bus implementation for use with asyncio.

    The message bus class is the entry point into all the features of the
    library. It sets up a connection to the DBus daemon and exposes an
    interface to send and receive messages and expose services.

    You must call :func:`connect() <dbus_fast.aio.MessageBus.connect>` before
    using this message bus.

    :param bus_type: The type of bus to connect to. Affects the search path for
        the bus address.
    :type bus_type: :class:`BusType <dbus_fast.BusType>`
    :param bus_address: A specific bus address to connect to. Should not be
        used under normal circumstances.
    :param auth: The authenticator to use, defaults to an instance of
        :class:`AuthExternal <dbus_fast.auth.AuthExternal>`.
    :type auth: :class:`Authenticator <dbus_fast.auth.Authenticator>`
    :param negotiate_unix_fd: Allow the bus to send and receive Unix file
        descriptors (DBus type 'h'). This must be supported by the transport.
    :type negotiate_unix_fd: bool

    :ivar unique_name: The unique name of the message bus connection. It will
        be :class:`None` until the message bus connects.
    :vartype unique_name: str
    :ivar connected: True if this message bus is expected to be able to send
        and receive messages.
    :vartype connected: bool
    """

    __slots__ = ("_auth", "_disconnect_future", "_loop", "_pending_futures", "_writer")

    def __init__(
        self,
        bus_address: str | None = None,
        bus_type: BusType = BusType.SESSION,
        auth: Authenticator | None = None,
        negotiate_unix_fd: bool = False,
    ) -> None:
        super().__init__(bus_address, bus_type, ProxyObject, negotiate_unix_fd)
        self._loop = asyncio.get_running_loop()

        self._writer = _MessageWriter(self)

        if auth is None:
            self._auth = AuthExternal()
        else:
            self._auth = auth

        self._disconnect_future = self._loop.create_future()
        self._pending_futures: set[asyncio.Future] = set()

    async def connect(self) -> MessageBus:
        """Connect this message bus to the DBus daemon.

        This method must be called before the message bus can be used.

        :returns: This message bus for convenience.
        :rtype: :class:`MessageBus <dbus_fast.aio.MessageBus>`

        :raises:
            - :class:`AuthError <dbus_fast.AuthError>` - If authorization to \
              the DBus daemon failed.
            - :class:`Exception` - If there was a connection error.
        """
        await self._authenticate()

        future = self._loop.create_future()

        self._loop.add_reader(
            self._fd,
            build_message_reader(
                self._sock,
                self._process_message,
                self._finalize,
                self._negotiate_unix_fd,
            ),
        )

        def on_hello(reply, err):
            try:
                if err:
                    raise err
                self.unique_name = reply.body[0]
                self._writer.schedule_write()
                _future_set_result(future, self)
            except Exception as e:
                _future_set_exception(future, e)
                self.disconnect()
                self._finalize(err)

        next_serial = self.next_serial()
        self._method_return_handlers[next_serial] = on_hello
        if next_serial == 1:
            serialized = HELLO_1_SERIALIZED
        else:
            serialized = _generate_hello_serialized(next_serial)
        self._stream.write(serialized)
        self._stream.flush()
        return await future

    async def introspect(
        self,
        bus_name: str,
        path: str,
        timeout: float = 30.0,
        validate_property_names: bool = True,
    ) -> intr.Node:
        """Get introspection data for the node at the given path from the given
        bus name.

        Calls the standard ``org.freedesktop.DBus.Introspectable.Introspect``
        on the bus for the path.

        :param bus_name: The name to introspect.
        :type bus_name: str
        :param path: The path to introspect.
        :type path: str
        :param timeout: The timeout to introspect.
        :type timeout: float
        :param validate_property_names: Whether to validate property names or not.
        :type validate_property_names: bool

        :returns: The introspection data for the name at the path.
        :rtype: :class:`Node <dbus_fast.introspection.Node>`

        :raises:
            - :class:`InvalidObjectPathError <dbus_fast.InvalidObjectPathError>` \
                    - If the given object path is not valid.
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
            - :class:`asyncio.TimeoutError` - Waited for future but time run out.
        """
        future = self._loop.create_future()

        super().introspect(
            bus_name,
            path,
            partial(self._reply_handler, future),
            check_callback_type=False,
            validate_property_names=validate_property_names,
        )

        timer_handle = self._loop.call_later(
            timeout, _future_set_exception, future, asyncio.TimeoutError
        )
        try:
            return await future
        finally:
            timer_handle.cancel()

    async def request_name(
        self, name: str, flags: NameFlag = NameFlag.NONE
    ) -> RequestNameReply:
        """Request that this message bus owns the given name.

        :param name: The name to request.
        :type name: str
        :param flags: Name flags that affect the behavior of the name request.
        :type flags: :class:`NameFlag <dbus_fast.NameFlag>`

        :returns: The reply to the name request.
        :rtype: :class:`RequestNameReply <dbus_fast.RequestNameReply>`

        :raises:
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        future = self._loop.create_future()

        super().request_name(
            name, flags, partial(self._reply_handler, future), check_callback_type=False
        )

        return await future

    async def release_name(self, name: str) -> ReleaseNameReply:
        """Request that this message bus release the given name.

        :param name: The name to release.
        :type name: str

        :returns: The reply to the release request.
        :rtype: :class:`ReleaseNameReply <dbus_fast.ReleaseNameReply>`

        :raises:
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        future = self._loop.create_future()

        super().release_name(
            name, partial(self._reply_handler, future), check_callback_type=False
        )

        return await future

    async def call(self, msg: Message) -> Message:
        """Send a method call and wait for a reply from the DBus daemon.

        :param msg: The method call message to send.
        :type msg: :class:`Message <dbus_fast.Message>`

        :returns: A message in reply to the method call sent.
        :rtype: :class:`Message <dbus_fast.Message>`

        :raises:
            - :class:`Exception` - If a connection error occurred.

        .. versionchanged:: 3.1.0
            Using this to call methods that do not expect a reply or messages
            that are not a method call is deprecated. Use :func:`send()
            <dbus_fast.aio.MessageBus.send>` instead for those cases.
        """
        if (
            msg.flags.value & NO_REPLY_EXPECTED_VALUE
            or msg.message_type is not MessageType.METHOD_CALL
        ):
            warn(
                "Using this to call methods that do not expect a reply or messages that are not a method call is deprecated. Use send() instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            await self.send(msg)
            return None

        future = self._loop.create_future()

        self._call(msg, partial(self._reply_handler, future))

        await future

        return future.result()

    def send(self, msg: Message) -> asyncio.Future:
        """Asynchronously send a message on the message bus.

        .. note:: This method may change to a couroutine function in the 1.0
            release of the library.

        :param msg: The message to send.
        :type msg: :class:`Message <dbus_fast.Message>`

        :returns: A future that resolves when the message is sent or a
            connection error occurs.
        :rtype: :class:`Future <asyncio.Future>`
        """
        if not msg.serial:
            msg.serial = self.next_serial()

        future = self._loop.create_future()
        self._writer.schedule_write(msg, future)
        return future

    def get_proxy_object(
        self, bus_name: str, path: str, introspection: intr.Node
    ) -> ProxyObject:
        return super().get_proxy_object(bus_name, path, introspection)

    async def wait_for_disconnect(self):
        """Wait for the message bus to disconnect.

        :returns: :class:`None` when the message bus has disconnected.
        :rtype: :class:`None`

        :raises:
            - :class:`Exception` - If connection was terminated unexpectedly or \
              an internal error occurred in the library.
        """
        return await self._disconnect_future

    def _future_exception_no_reply(self, fut: asyncio.Future) -> None:
        """Log an exception from a future that was not expected."""
        self._pending_futures.discard(fut)
        try:
            fut.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            _LOGGER.exception("unexpected exception in future", exc_info=e)

    def _make_method_handler(
        self, interface: ServiceInterface, method: _Method
    ) -> Callable[[Message, Callable[[Message], None]], None]:
        if not inspect.iscoroutinefunction(method.fn):
            return super()._make_method_handler(interface, method)

        negotiate_unix_fd = self._negotiate_unix_fd
        msg_body_to_args = ServiceInterface._msg_body_to_args
        fn_result_to_body = ServiceInterface._fn_result_to_body

        def _coroutine_method_handler(
            msg: Message, send_reply: Callable[[Message], None]
        ) -> None:
            """A coroutine method handler."""
            args = msg_body_to_args(msg) if msg.unix_fds else msg.body
            fut: asyncio.Future = asyncio.ensure_future(method.fn(interface, *args))
            # Hold a strong reference to the future to ensure
            # it is not garbage collected before it is done.
            self._pending_futures.add(fut)
            if (
                send_reply is _block_unexpected_reply
                or msg.flags.value & NO_REPLY_EXPECTED_VALUE
            ):
                fut.add_done_callback(self._future_exception_no_reply)
                return

            # We only create the closure function if we are actually going to reply
            def _done(fut: asyncio.Future) -> None:
                """The callback for when the method is done."""
                with send_reply:
                    result = fut.result()
                    body, unix_fds = fn_result_to_body(
                        result, method.out_signature_tree, replace_fds=negotiate_unix_fd
                    )
                    send_reply(
                        Message.new_method_return(
                            msg, method.out_signature, body, unix_fds
                        )
                    )

            fut.add_done_callback(_done)
            # Discard the future only after running the done callback
            fut.add_done_callback(self._pending_futures.discard)

        return _coroutine_method_handler

    async def _auth_readline(self) -> str:
        buf = b""
        while buf[-2:] != b"\r\n":
            chunk = await self._loop.sock_recv(self._sock, _AUTH_READ_CHUNK)
            if not chunk:
                raise AuthError("connection closed during authentication")
            buf += chunk
            if len(buf) > _MAX_AUTH_LINE:
                raise AuthError("auth line exceeded maximum size")
        return buf[:-2].decode()

    async def _authenticate(self) -> None:
        await self._loop.sock_sendall(self._sock, b"\0")

        first_line = self._auth._authentication_start(
            negotiate_unix_fd=self._negotiate_unix_fd
        )

        if first_line is not None:
            if type(first_line) is not str:
                raise AuthError("authenticator gave response not type str")
            await self._loop.sock_sendall(
                self._sock, Authenticator._format_line(first_line)
            )

        while True:
            response = self._auth._receive_line(await self._auth_readline())
            if response is not None:
                await self._loop.sock_sendall(
                    self._sock, Authenticator._format_line(response)
                )
                self._stream.flush()
            if response == "BEGIN":
                # The first octet received by the server after the \r\n of the BEGIN command
                # from the client must be the first octet of the authenticated/encrypted stream
                # of D-Bus messages.
                break

    def _finalize(self, err: Exception | None = None) -> None:
        try:
            self._loop.remove_reader(self._fd)
        except Exception:
            _LOGGER.warning("could not remove message reader", exc_info=True)
        try:
            self._loop.remove_writer(self._fd)
        except Exception:
            _LOGGER.warning("could not remove message writer", exc_info=True)

        had_handlers = bool(self._method_return_handlers or self._user_message_handlers)

        super()._finalize(err)

        if self._disconnect_future.done():
            return

        if err and not self._user_disconnect:
            _future_set_exception(self._disconnect_future, err)
            # If this happens during a reply, the message handlers
            # will have the exception set and wait_for_disconnect will
            # never be called so we need to manually set the exception
            # as retrieved to avoid asyncio warnings when the future
            # is garbage collected.
            if had_handlers:
                with contextlib.suppress(Exception):
                    self._disconnect_future.exception()
        else:
            _future_set_result(self._disconnect_future, None)

    def _reply_handler(
        self, future: asyncio.Future, reply: Any | None, err: Exception | None
    ) -> None:
        """The reply handler for method calls."""
        if err:
            _future_set_exception(future, err)
        else:
            _future_set_result(future, reply)



from dbus_fast.constants import MESSAGE_FLAG_MAP, ErrorType, MessageFlag
from dbus_fast.errors import DBusError


def test_message_flag_map():
    assert 0 in MESSAGE_FLAG_MAP
    assert MessageFlag.NONE in MESSAGE_FLAG_MAP


def test_error_type():
    err = DBusError(ErrorType.FAILED, "")
    assert err.type == ErrorType.FAILED



#!/usr/bin/env python3
import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__) + "/.."))

import asyncio
import json

from dbus_fast import Message, MessageType
from dbus_fast.aio import MessageBus


async def main():
    bus = await MessageBus().connect()

    reply = await bus.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="ListNames",
        )
    )

    if reply.message_type == MessageType.ERROR:
        raise Exception(reply.body[0])

    print(json.dumps(reply.body[0], indent=2))


asyncio.run(main())



import io
import timeit

from dbus_fast._private.unmarshaller import Unmarshaller

bluez_passive_message = (
    b"l\1\1\1*\0\0\0\205D\267\3\215\0\0\0\1\1o\0\35\0\0\0/org/bleak/61/281472597302272\0\0\0\6\1s\0\7\0\0"
    b"\0:1.1450\0\2\1s\0\37\0\0\0org.bluez.AdvertisementMonitor1\0\3\1s\0\v\0\0\0DeviceFound\0\0\0\0\0\10"
    b"\1g\0\1o\0\0\7\1s\0\4\0\0\0:1.4\0\0\0\0%\0\0\0/org/bluez/hci0/dev_58_D3_49_E6_02_6E\0l\1\1\1*\0\0\0"
    b"\206D\267\3\215\0\0\0\1\1o\0\35\0\0\0/org/bleak/61/281472593362560\0\0\0\6\1s\0\7\0\0\0:1.1450\0\2"
    b"\1s\0\37\0\0\0org.bluez.AdvertisementMonitor1\0\3\1s\0\v\0\0\0DeviceFound\0\0\0\0\0\10\1g\0\1o\0\0"
    b"\7\1s\0\4\0\0\0:1.4\0\0\0\0%\0\0\0/org/bluez/hci1/dev_58_D3_49_E6_02_6E\0"
)


stream = io.BytesIO(bluez_passive_message)

unmarshaller = Unmarshaller(stream)


def unmarhsall_bluez_rssi_message():
    stream.seek(0)
    unmarshaller.unmarshall()
    unmarshaller.unmarshall()


count = 3000000
time = timeit.Timer(unmarhsall_bluez_rssi_message).timeit(count)
print(f"Unmarshalling {count} bluetooth passive messages took {time} seconds")



#!/usr/bin/env python3
import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__) + "/.."))

import asyncio
from typing import Annotated

from dbus_fast import Variant
from dbus_fast.aio.message_bus import MessageBus
from dbus_fast.annotations import DBusDict, DBusSignature, DBusStr
from dbus_fast.service import ServiceInterface, dbus_method, dbus_property, dbus_signal


class ExampleInterface(ServiceInterface):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self._string_prop = "kevin"

    @dbus_method()
    def Echo(self, what: DBusStr) -> DBusStr:
        return what

    @dbus_method()
    def EchoMultiple(
        self, what1: DBusStr, what2: DBusStr
    ) -> Annotated[tuple[str, str], DBusSignature("ss")]:
        return what1, what2

    @dbus_method()
    def GetVariantDict(self) -> DBusDict:
        return {
            "foo": Variant("s", "bar"),
            "bat": Variant("x", -55),
            "a_list": Variant("as", ["hello", "world"]),
        }

    @dbus_property(name="StringProp")
    def string_prop(self) -> DBusStr:
        return self._string_prop

    @string_prop.setter
    def string_prop_setter(self, val: DBusStr) -> None:
        self._string_prop = val

    @dbus_signal()
    def signal_simple(self) -> DBusStr:
        return "hello"

    @dbus_signal()
    def signal_multiple(
        self,
    ) -> Annotated[tuple[str, str], DBusSignature("ss")]:
        return "hello", "world"


async def main() -> None:
    name = "dbus.next.example.service"
    path = "/example/path"
    interface_name = "example.interface"

    bus = await MessageBus().connect()
    interface = ExampleInterface(interface_name)
    bus.export("/example/path", interface)
    await bus.request_name(name)
    print(
        f'service up on name: "{name}", path: "{path}", interface: "{interface_name}"'
    )
    await bus.wait_for_disconnect()


asyncio.run(main())



# cython: freethreading_compatible = True

from __future__ import annotations

import array
import errno
import io
import socket
import sys
from collections.abc import Callable, Iterable
from struct import Struct
from typing import TYPE_CHECKING, Any

from ..constants import MESSAGE_FLAG_MAP, MESSAGE_TYPE_MAP, MessageFlag
from ..errors import InvalidMessageError
from ..message import Message
from ..signature import SignatureType, Variant, get_signature_tree
from .constants import BIG_ENDIAN, LITTLE_ENDIAN, PROTOCOL_VERSION

MESSAGE_FLAG_INTENUM = MessageFlag

MAX_UNIX_FDS = 16
MAX_UNIX_FDS_SIZE = array.array("i").itemsize
UNIX_FDS_CMSG_LENGTH = socket.CMSG_LEN(MAX_UNIX_FDS_SIZE * MAX_UNIX_FDS)

UNPACK_SYMBOL = {LITTLE_ENDIAN: "<", BIG_ENDIAN: ">"}

UINT32_CAST = "I"
UINT32_SIZE = 4
UINT32_DBUS_TYPE = "u"

INT16_CAST = "h"
INT16_SIZE = 2
INT16_DBUS_TYPE = "n"

UINT16_CAST = "H"
UINT16_SIZE = 2
UINT16_DBUS_TYPE = "q"

SYS_IS_LITTLE_ENDIAN = sys.byteorder == "little"
SYS_IS_BIG_ENDIAN = sys.byteorder == "big"

DBUS_TO_CTYPE = {
    "y": ("B", 1),  # byte
    INT16_DBUS_TYPE: (INT16_CAST, INT16_SIZE),  # int16
    UINT16_DBUS_TYPE: (UINT16_CAST, UINT16_SIZE),  # uint16
    "i": ("i", 4),  # int32
    UINT32_DBUS_TYPE: (UINT32_CAST, UINT32_SIZE),  # uint32
    "x": ("q", 8),  # int64
    "t": ("Q", 8),  # uint64
    "d": ("d", 8),  # double
    "h": (UINT32_CAST, UINT32_SIZE),  # uint32
}

UNPACK_HEADER_LITTLE_ENDIAN = Struct("<III").unpack_from
UNPACK_HEADER_BIG_ENDIAN = Struct(">III").unpack_from

UINT32_UNPACK_LITTLE_ENDIAN = Struct(f"<{UINT32_CAST}").unpack_from
UINT32_UNPACK_BIG_ENDIAN = Struct(f">{UINT32_CAST}").unpack_from

INT16_UNPACK_LITTLE_ENDIAN = Struct(f"<{INT16_CAST}").unpack_from
INT16_UNPACK_BIG_ENDIAN = Struct(f">{INT16_CAST}").unpack_from

UINT16_UNPACK_LITTLE_ENDIAN = Struct(f"<{UINT16_CAST}").unpack_from
UINT16_UNPACK_BIG_ENDIAN = Struct(f">{UINT16_CAST}").unpack_from

HEADER_SIGNATURE_SIZE = 16
HEADER_ARRAY_OF_STRUCT_SIGNATURE_POSITION = 12

# D-Bus spec: total message length must not exceed 128 MiB. Enforced on the
# attacker-controlled header_len / body_len fields before any allocation or
# socket read sized by them, so a forged header can't push the process into
# an OOM. https://dbus.freedesktop.org/doc/dbus-specification.html
#
# The check in _read_header bounds header_len and body_len independently
# (each <= 128 MiB, sum <= 128 MiB). The actual on-wire read in _read_body
# is HEADER_SIGNATURE_SIZE + _msg_len, where _msg_len adds up to 7 bytes of
# header-to-body alignment padding — so the worst-case read is the cap plus
# HEADER_SIGNATURE_SIZE (16) + 7 = 23 bytes of slack. That's deliberate: the
# DoS vector is "buffer gigabytes from a forged header", and 23 bytes over
# the spec ceiling doesn't change that. Keeping the check on the raw fields
# is simpler than computing the padded total here.
#
# MAX_MESSAGE_SIZE is the Python-importable form (used by tests);
# _MAX_MESSAGE_SIZE is the cdef unsigned int form used internally per .pxd.
MAX_MESSAGE_SIZE = 134_217_728
_MAX_MESSAGE_SIZE = MAX_MESSAGE_SIZE


# Most common signatures

SIGNATURE_TREE_EMPTY = get_signature_tree("")
SIGNATURE_TREE_B = get_signature_tree("b")
SIGNATURE_TREE_N = get_signature_tree("n")
SIGNATURE_TREE_S = get_signature_tree("s")
SIGNATURE_TREE_O = get_signature_tree("o")
SIGNATURE_TREE_U = get_signature_tree("u")
SIGNATURE_TREE_Y = get_signature_tree("y")

SIGNATURE_TREE_AY = get_signature_tree("ay")
SIGNATURE_TREE_AS = get_signature_tree("as")
SIGNATURE_TREE_AS_TYPES_0 = SIGNATURE_TREE_AS.root_type
SIGNATURE_TREE_A_SV = get_signature_tree("a{sv}")
SIGNATURE_TREE_A_SV_TYPES_0 = SIGNATURE_TREE_A_SV.root_type

SIGNATURE_TREE_AO = get_signature_tree("ao")
SIGNATURE_TREE_AO_TYPES_0 = SIGNATURE_TREE_AO.root_type

SIGNATURE_TREE_OAS = get_signature_tree("oas")
SIGNATURE_TREE_OAS_TYPES_1 = SIGNATURE_TREE_OAS.types[1]

SIGNATURE_TREE_AY_TYPES_0 = SIGNATURE_TREE_AY.root_type
SIGNATURE_TREE_A_QV = get_signature_tree("a{qv}")
SIGNATURE_TREE_A_QV_TYPES_0 = SIGNATURE_TREE_A_QV.root_type

SIGNATURE_TREE_SA_SV_AS = get_signature_tree("sa{sv}as")
SIGNATURE_TREE_SA_SV_AS_TYPES_1 = SIGNATURE_TREE_SA_SV_AS.types[1]
SIGNATURE_TREE_SA_SV_AS_TYPES_2 = SIGNATURE_TREE_SA_SV_AS.types[2]

SIGNATURE_TREE_OA_SA_SV = get_signature_tree("oa{sa{sv}}")
SIGNATURE_TREE_OA_SA_SV_TYPES_1 = SIGNATURE_TREE_OA_SA_SV.types[1]

SIGNATURE_TREE_A_OA_SA_SV = get_signature_tree("a{oa{sa{sv}}}")
SIGNATURE_TREE_A_OA_SA_SV_TYPES_0 = SIGNATURE_TREE_A_OA_SA_SV.root_type


TOKEN_B_AS_INT = ord("b")
TOKEN_U_AS_INT = ord("u")
TOKEN_Y_AS_INT = ord("y")
TOKEN_A_AS_INT = ord("a")
TOKEN_O_AS_INT = ord("o")
TOKEN_S_AS_INT = ord("s")
TOKEN_G_AS_INT = ord("g")
TOKEN_N_AS_INT = ord("n")
TOKEN_X_AS_INT = ord("x")
TOKEN_T_AS_INT = ord("t")
TOKEN_D_AS_INT = ord("d")
TOKEN_Q_AS_INT = ord("q")
TOKEN_V_AS_INT = ord("v")
TOKEN_LEFT_CURLY_AS_INT = ord("{")
TOKEN_LEFT_PAREN_AS_INT = ord("(")


VARIANT_BOOL_TRUE = Variant._factory(SIGNATURE_TREE_B, True)
VARIANT_BOOL_FALSE = Variant._factory(SIGNATURE_TREE_B, False)


ARRAY = array.array
SOL_SOCKET = socket.SOL_SOCKET
SCM_RIGHTS = socket.SCM_RIGHTS

EAGAIN = errno.EAGAIN
EWOULDBLOCK = errno.EWOULDBLOCK

HEADER_IDX_TO_ARG_NAME = [
    "",
    "path",
    "interface",
    "member",
    "error_name",
    "reply_serial",
    "destination",
    "sender",
    "signature",
    "unix_fds",
]
HEADER_PATH_IDX = HEADER_IDX_TO_ARG_NAME.index("path")
HEADER_INTERFACE_IDX = HEADER_IDX_TO_ARG_NAME.index("interface")
HEADER_MEMBER_IDX = HEADER_IDX_TO_ARG_NAME.index("member")
HEADER_ERROR_NAME_IDX = HEADER_IDX_TO_ARG_NAME.index("error_name")
HEADER_REPLY_SERIAL_IDX = HEADER_IDX_TO_ARG_NAME.index("reply_serial")
HEADER_DESTINATION_IDX = HEADER_IDX_TO_ARG_NAME.index("destination")
HEADER_SENDER_IDX = HEADER_IDX_TO_ARG_NAME.index("sender")
HEADER_SIGNATURE_IDX = HEADER_IDX_TO_ARG_NAME.index("signature")
HEADER_UNIX_FDS_IDX = HEADER_IDX_TO_ARG_NAME.index("unix_fds")

_EMPTY_HEADERS: list[Any | None] = [None] * len(HEADER_IDX_TO_ARG_NAME)

_SignatureType = SignatureType
_int = int

READER_TYPE = Callable[["Unmarshaller", SignatureType], Any]

MARSHALL_STREAM_END_ERROR = BlockingIOError

DEFAULT_BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE


def unpack_parser_factory(unpack_from: Callable, size: int) -> READER_TYPE:
    """Build a parser that unpacks the bytes using the given unpack_from function."""

    def _unpack_from_parser(self: Unmarshaller, signature: SignatureType) -> Any:
        self._pos += size + (-self._pos & (size - 1))  # align
        return unpack_from(self._buf, self._pos - size)[0]

    return _unpack_from_parser


def build_simple_parsers(
    endian: int,
) -> dict[str, Callable[[Unmarshaller, SignatureType], Any]]:
    """Build a dict of parsers for simple types."""
    parsers: dict[str, READER_TYPE] = {}
    for dbus_type, ctype_size in DBUS_TO_CTYPE.items():
        ctype, size = ctype_size
        size = ctype_size[1]
        parsers[dbus_type] = unpack_parser_factory(
            Struct(f"{UNPACK_SYMBOL[endian]}{ctype}").unpack_from, size
        )
    return parsers


try:
    import cython
except ImportError:
    from ._cython_compat import FAKE_CYTHON as cython
int_ = int
bytearray_ = bytearray


def is_compiled() -> bool:
    return cython.compiled


def _ustr_uint32(buf: bytearray_, pos: int_, endian: int_) -> int_:
    if endian == LITTLE_ENDIAN:
        return (
            buf[pos] | (buf[pos + 1] << 8) | (buf[pos + 2] << 16) | (buf[pos + 3] << 24)
        )
    return buf[pos + 3] | (buf[pos + 2] << 8) | (buf[pos + 1] << 16) | (buf[pos] << 24)


def buffer_to_uint32(buf: bytearray, pos: int, endian: int) -> int:
    return _ustr_uint32(buf, pos, endian)


def _ustr_int16(buf: bytearray_, pos: int_, endian: int_) -> int_:
    # Caution: this function will only work with Cython
    # because it relies on casting the result to a signed int
    # and will return an unsigned int if not compiled.
    if endian == LITTLE_ENDIAN:
        return buf[pos] | (buf[pos + 1] << 8)  # pragma: no cover
    return buf[pos + 1] | (buf[pos] << 8)  # pragma: no cover


def buffer_to_int16(buf: bytearray | bytes, pos: int, endian: int) -> int:
    # Caution: this function will only work with Cython
    # because it relies on casting the result to a signed int
    # and will return an unsigned int if not compiled.
    return _ustr_int16(buf, pos, endian)


def _ustr_uint16(buf: bytearray_, pos: int_, endian: int_) -> int_:
    if endian == LITTLE_ENDIAN:
        return buf[pos] | (buf[pos + 1] << 8)
    return buf[pos + 1] | (buf[pos] << 8)


def buffer_to_uint16(buf: bytearray, pos: int, endian: int) -> int:
    return _ustr_uint16(buf, pos, endian)


# Alignment padding is handled with the following formula below
#
# For any align value, the correct padding formula is:
#
#    (align - (pos % align)) % align
#
# However, if align is a power of 2 (always the case here), the slow MOD
# operator can be replaced by a bitwise AND:
#
#    (align - (pos & (align - 1))) & (align - 1)
#
# Which can be simplified to:
#
#    (-pos) & (align - 1)
#
#
class Unmarshaller:
    """Unmarshall messages from a stream.

    When calling with sock and _negotiate_unix_fd False, the unmashaller must
    be called continuously for each new message as it will buffer the data
    until a complete message is available.
    """

    __slots__ = (
        "_body_len",
        "_buf",
        "_buf_len",
        "_buf_ustr",
        "_endian",
        "_flag",
        "_header_len",
        "_int16_unpack",
        "_message",
        "_message_type",
        "_msg_len",
        "_negotiate_unix_fd",
        "_pos",
        "_read_complete",
        "_readers",
        "_serial",
        "_sock",
        "_sock_with_fds_reader",
        "_sock_without_fds_reader",
        "_stream",
        "_stream_reader",
        "_uint16_unpack",
        "_uint32_unpack",
        "_unix_fds",
    )

    _stream_reader: Callable[[int], bytes]

    def __init__(
        self,
        stream: io.BufferedRWPair | None = None,
        sock: socket.socket | None = None,
        negotiate_unix_fd: bool = True,
    ) -> None:
        self._unix_fds: list[int] = []
        self._buf: bytearray = bytearray()  # Actual buffer
        self._buf_ustr = self._buf  # Used to avoid type checks
        self._buf_len = 0
        self._stream = stream
        self._sock = sock
        self._message: Message | None = None
        self._readers: dict[str, READER_TYPE] = {}
        self._pos = 0
        self._body_len = 0
        self._serial = 0
        self._header_len = 0
        self._message_type = 0
        self._flag = 0
        self._msg_len = 0
        self._uint32_unpack: Callable[[bytearray, int], tuple[int]] | None = None
        self._int16_unpack: Callable[[bytearray, int], tuple[int]] | None = None
        self._uint16_unpack: Callable[[bytearray, int], tuple[int]] | None = None
        self._negotiate_unix_fd = negotiate_unix_fd
        self._read_complete = False
        if stream:
            if isinstance(stream, io.BufferedRWPair) and hasattr(stream, "reader"):
                self._stream_reader = stream.reader.read
            else:
                self._stream_reader = stream.read
        elif self._negotiate_unix_fd:
            if TYPE_CHECKING:
                assert self._sock is not None
            self._sock_with_fds_reader = self._sock.recvmsg
        else:
            if TYPE_CHECKING:
                assert self._sock is not None
            self._sock_without_fds_reader = self._sock.recv
        self._endian = 0

    def _next_message(self) -> None:
        """Reset the unmarshaller to its initial state.

        Call this before processing a new message.
        """
        if self._unix_fds:
            self._unix_fds = []
        to_clear = HEADER_SIGNATURE_SIZE + self._msg_len
        if self._buf_len == to_clear:
            del self._buf[:]
            self._buf_len = 0
        else:
            del self._buf[:to_clear]
            self._buf_len -= to_clear
            self._buf_ustr = self._buf
        self._msg_len = 0  # used to check if we have ready the header
        self._read_complete = False  # used to check if we have ready the message
        # No need to reset the unpack functions, they are set in _read_header
        # every time a new message is processed.

    @property
    def message(self) -> Message | None:
        """Return the message that has been unmarshalled."""
        if self._read_complete:
            return self._message
        return None

    def _has_another_message_in_buffer(self) -> bool:
        """Check if there is another message in the buffer."""
        return self._buf_len > HEADER_SIGNATURE_SIZE + self._msg_len

    def _read_sock_with_fds(self, pos: _int, missing_bytes: _int) -> None:
        """reads from the socket, storing any fds sent and handling errors
        from the read itself.

        This function is greedy and will read as much data as possible
        from the underlying socket.
        """
        # This will raise BlockingIOError if there is no data to read
        # which we store in the MARSHALL_STREAM_END_ERROR object
        try:
            recv = self._sock_with_fds_reader(missing_bytes, UNIX_FDS_CMSG_LENGTH)
        except OSError as e:
            errno = e.errno
            if errno == EAGAIN or errno == EWOULDBLOCK:
                raise MARSHALL_STREAM_END_ERROR
            raise
        msg = recv[0]
        ancdata = recv[1]
        if ancdata:
            for level, type_, data in ancdata:
                if not (level == SOL_SOCKET and type_ == SCM_RIGHTS):
                    continue
                self._unix_fds.extend(
                    ARRAY("i", data[: len(data) - (len(data) % MAX_UNIX_FDS_SIZE)])
                )
        if not msg:
            raise EOFError
        self._buf += msg
        self._buf_len = len(self._buf)
        if self._buf_len < pos:
            raise MARSHALL_STREAM_END_ERROR

    def _read_sock_without_fds(self, pos: _int) -> None:
        """reads from the socket and handling errors from the read itself.

        This function is greedy and will read as much data as possible
        from the underlying socket.
        """
        # This will raise BlockingIOError if there is no data to read
        # which we store in the MARSHALL_STREAM_END_ERROR object
        while True:
            try:
                data = self._sock_without_fds_reader(DEFAULT_BUFFER_SIZE)
            except OSError as e:
                errno = e.errno
                if errno == EAGAIN or errno == EWOULDBLOCK:
                    raise MARSHALL_STREAM_END_ERROR
                raise
            if not data:
                raise EOFError
            self._buf += data
            self._buf_len = len(self._buf)
            if self._buf_len >= pos:
                return

    def _read_stream(self, pos: _int, missing_bytes: _int) -> None:
        """Read from the stream."""
        data = self._stream_reader(missing_bytes)
        if data is None:
            raise MARSHALL_STREAM_END_ERROR
        if not data:
            raise EOFError
        self._buf += data
        self._buf_len = len(self._buf)
        if self._buf_len < pos:
            raise MARSHALL_STREAM_END_ERROR

    def _read_to_pos(self, pos: _int) -> None:
        """
        Read from underlying socket into buffer.

        Raises BlockingIOError if there is not enough data to be read.

        :arg pos:
            The pos to read to. If not enough bytes are available in the
            buffer, read more from it.

        :returns:
            None
        """
        missing_bytes = pos - self._buf_len
        if missing_bytes <= 0:
            return
        if self._sock is None:
            self._read_stream(pos, missing_bytes)
        elif self._negotiate_unix_fd:
            self._read_sock_with_fds(pos, missing_bytes)
        else:
            self._read_sock_without_fds(pos)
        self._buf_ustr = self._buf

    def read_uint32_unpack(self, type_: _SignatureType) -> int:
        return self._read_uint32_unpack()

    def _read_uint32_unpack(self) -> int:
        self._pos += UINT32_SIZE + (-self._pos & (UINT32_SIZE - 1))  # align
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read uint32")
            return _ustr_uint32(self._buf_ustr, self._pos - UINT32_SIZE, self._endian)
        return self._uint32_unpack(self._buf, self._pos - UINT32_SIZE)[0]

    def read_uint16_unpack(self, type_: _SignatureType) -> int:
        return self._read_uint16_unpack()

    def _read_uint16_unpack(self) -> int:
        self._pos += UINT16_SIZE + (-self._pos & (UINT16_SIZE - 1))  # align
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read uint16")
            return _ustr_uint16(self._buf_ustr, self._pos - UINT16_SIZE, self._endian)
        return self._uint16_unpack(self._buf, self._pos - UINT16_SIZE)[0]

    def read_int16_unpack(self, type_: _SignatureType) -> int:
        return self._read_int16_unpack()

    def _read_int16_unpack(self) -> int:
        self._pos += INT16_SIZE + (-self._pos & (INT16_SIZE - 1))  # align
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read int16")
            return _ustr_int16(self._buf_ustr, self._pos - INT16_SIZE, self._endian)
        return self._int16_unpack(self._buf, self._pos - INT16_SIZE)[0]

    def read_boolean(self, type_: _SignatureType) -> bool:
        return self._read_boolean()

    def _read_boolean(self) -> bool:
        return bool(self._read_uint32_unpack())

    def read_string_unpack(self, type_: _SignatureType) -> str:
        return self._read_string_unpack()

    def _read_string_unpack(self) -> str:
        """Read a string using unpack."""
        self._pos += UINT32_SIZE + (-self._pos & (UINT32_SIZE - 1))  # align
        str_start = self._pos
        # read terminating '\0' byte as well (str_length + 1)
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read uint32")
            self._pos += (
                _ustr_uint32(self._buf_ustr, str_start - UINT32_SIZE, self._endian) + 1
            )
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read string")
        else:
            self._pos += self._uint32_unpack(self._buf, str_start - UINT32_SIZE)[0] + 1
        return self._buf_ustr[str_start : self._pos - 1].decode()

    def read_signature(self, type_: _SignatureType) -> str:
        return self._read_signature()

    def _read_signature(self) -> str:
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read signature")
        signature_len = self._buf_ustr[self._pos]  # byte
        o = self._pos + 1
        # read terminating '\0' byte as well (str_length + 1)
        self._pos = o + signature_len + 1
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read signature")
        return self._buf_ustr[o : o + signature_len].decode()

    def read_variant(self, type_: _SignatureType) -> Variant:
        return self._read_variant()

    def _read_variant(self) -> Variant:
        # Inline signature reading to avoid _read_signature() call,
        # .decode(), and ord() for the common single-byte case.
        # verify in Variant is only useful on construction not unmarshalling
        if cython.compiled:
            if self._buf_len <= self._pos:
                raise IndexError("Not enough data to read signature")
        signature_len = self._buf_ustr[self._pos]
        if signature_len == 1:
            self._pos += 3  # 1 len byte + 1 signature char + 1 null terminator
            if cython.compiled:
                if self._buf_len < self._pos:
                    raise IndexError("Not enough data to read signature")
            token_as_int = self._buf_ustr[self._pos - 2]
            if token_as_int == TOKEN_N_AS_INT:
                return Variant._factory(SIGNATURE_TREE_N, self._read_int16_unpack())
            if token_as_int == TOKEN_S_AS_INT:
                return Variant._factory(SIGNATURE_TREE_S, self._read_string_unpack())
            if token_as_int == TOKEN_B_AS_INT:
                return VARIANT_BOOL_TRUE if self._read_boolean() else VARIANT_BOOL_FALSE
            if token_as_int == TOKEN_O_AS_INT:
                return Variant._factory(SIGNATURE_TREE_O, self._read_string_unpack())
            if token_as_int == TOKEN_U_AS_INT:
                return Variant._factory(SIGNATURE_TREE_U, self._read_uint32_unpack())
            if token_as_int == TOKEN_Y_AS_INT:
                self._pos += 1
                if cython.compiled:
                    if self._buf_len < self._pos:
                        raise IndexError("Not enough data to read byte")
                return Variant._factory(SIGNATURE_TREE_Y, self._buf_ustr[self._pos - 1])
            # Uncommon single-byte type, decode for fallback
            signature = chr(token_as_int)
        else:
            o = self._pos + 1
            self._pos = o + signature_len + 1
            if cython.compiled:
                if self._buf_len < self._pos:
                    raise IndexError("Not enough data to read signature")
            signature = self._buf_ustr[o : o + signature_len].decode()
            token_as_int = self._buf_ustr[o]
        if token_as_int == TOKEN_A_AS_INT:
            if signature == "ay":
                return Variant._factory(
                    SIGNATURE_TREE_AY, self.read_array(SIGNATURE_TREE_AY_TYPES_0)
                )
            if signature == "a{qv}":
                return Variant._factory(
                    SIGNATURE_TREE_A_QV, self.read_array(SIGNATURE_TREE_A_QV_TYPES_0)
                )
            if signature == "as":
                return Variant._factory(
                    SIGNATURE_TREE_AS, self.read_array(SIGNATURE_TREE_AS_TYPES_0)
                )
            if signature == "a{sv}":
                return Variant._factory(
                    SIGNATURE_TREE_A_SV, self.read_array(SIGNATURE_TREE_A_SV_TYPES_0)
                )
            if signature == "ao":
                return Variant._factory(
                    SIGNATURE_TREE_AO, self.read_array(SIGNATURE_TREE_AO_TYPES_0)
                )
        tree = get_signature_tree(signature)
        signature_type = tree.root_type
        return Variant._factory(
            tree, self._readers[signature_type.token](self, signature_type)
        )

    def read_struct(self, type_: _SignatureType) -> tuple[Any, ...]:
        self._pos += -self._pos & 7  # align 8
        readers = self._readers
        return tuple(
            readers[child_type.token](self, child_type) for child_type in type_.children
        )

    def read_dict_entry(self, type_: _SignatureType) -> tuple[Any, Any]:
        self._pos += -self._pos & 7  # align 8
        return self._readers[type_.children[0].token](
            self, type_.children[0]
        ), self._readers[type_.children[1].token](self, type_.children[1])

    def read_array(self, type_: _SignatureType) -> Iterable[Any]:
        self._pos += -self._pos & 3  # align 4 for the array
        self._pos += (
            -self._pos & (UINT32_SIZE - 1)
        ) + UINT32_SIZE  # align for the uint32
        if cython.compiled:
            if self._buf_len < self._pos:
                raise IndexError("Not enough data to read uint32")
            array_length = _ustr_uint32(
                self._buf_ustr, self._pos - UINT32_SIZE, self._endian
            )
        else:
            array_length = self._uint32_unpack(self._buf, self._pos - UINT32_SIZE)[0]
        child_type = type_._child_0
        token_as_int = child_type.token_as_int

        if token_as_int in {
            TOKEN_X_AS_INT,
            TOKEN_T_AS_INT,
            TOKEN_D_AS_INT,
            TOKEN_LEFT_CURLY_AS_INT,
            TOKEN_LEFT_PAREN_AS_INT,
        }:
            # the first alignment is not included in the array size
            self._pos += -self._pos & 7  # align 8

        if token_as_int == TOKEN_Y_AS_INT:
            self._pos += array_length
            if cython.compiled:
                if self._buf_len < self._pos:
                    raise IndexError("Not enough data to read byte")
            return self._buf_ustr[self._pos - array_length : self._pos]

        if token_as_int == TOKEN_LEFT_CURLY_AS_INT:
            result_dict: dict[Any, Any] = {}
            key: str | int
            beginning_pos = self._pos
            child_0 = child_type._child_0
            child_1 = child_type._child_1
            child_0_token_as_int = child_0.token_as_int
            child_1_token_as_int = child_1.token_as_int
            # Strings with variant values are the most common case
            # so we optimize for that by inlining the string reading
            # and the variant reading here
            if (
                child_0_token_as_int in {TOKEN_O_AS_INT, TOKEN_S_AS_INT}
                and child_1_token_as_int == TOKEN_V_AS_INT
            ):
                while self._pos - beginning_pos < array_length:
                    self._pos += -self._pos & 7  # align 8
                    key = self._read_string_unpack()
                    result_dict[key] = self._read_variant()
            elif (
                child_0_token_as_int == TOKEN_Q_AS_INT
                and child_1_token_as_int == TOKEN_V_AS_INT
            ):
                while self._pos - beginning_pos < array_length:
                    self._pos += -self._pos & 7  # align 8
                    key = self._read_uint16_unpack()
                    result_dict[key] = self._read_variant()
            elif (
                child_0_token_as_int in {TOKEN_O_AS_INT, TOKEN_S_AS_INT}
                and child_1_token_as_int == TOKEN_A_AS_INT
            ):
                while self._pos - beginning_pos < array_length:
                    self._pos += -self._pos & 7  # align 8
                    key = self._read_string_unpack()
                    result_dict[key] = self.read_array(child_1)
            else:
                reader_1 = self._readers[child_1.token]
                reader_0 = self._readers[child_0.token]
                while self._pos - beginning_pos < array_length:
                    self._pos += -self._pos & 7  # align 8
                    key = reader_0(self, child_0)
                    result_dict[key] = reader_1(self, child_1)

            return result_dict

        if array_length == 0:
            return []

        result_list = []
        beginning_pos = self._pos
        if token_as_int == TOKEN_O_AS_INT or token_as_int == TOKEN_S_AS_INT:
            while self._pos - beginning_pos < array_length:
                result_list.append(self._read_string_unpack())
            return result_list
        reader = self._readers[child_type.token]
        while self._pos - beginning_pos < array_length:
            result_list.append(reader(self, child_type))
        return result_list

    def _header_fields(self, header_length: _int) -> list[Any]:
        """Header fields are always a(yv)."""
        beginning_pos = self._pos
        headers = _EMPTY_HEADERS.copy()
        if cython.compiled:
            if self._buf_len < self._pos + header_length:
                raise IndexError("Not enough data to read header")
        while self._pos - beginning_pos < header_length:
            # Now read the y (byte) of struct (yv)
            self._pos += (-self._pos & 7) + 1  # align 8 + 1 for 'y' byte
            field_0 = self._buf_ustr[self._pos - 1]

            # Now read the v (variant) of struct (yv)
            # first we read the signature
            signature_len = self._buf_ustr[self._pos]  # byte
            o = self._pos + 1
            if cython.compiled:
                if self._buf_len < o + signature_len:
                    raise IndexError("Not enough data to read signature")
            self._pos += signature_len + 2  # one for the byte, one for the '\0'
            if field_0 == HEADER_UNIX_FDS_IDX:  # defined by self._unix_fds
                continue
            token_as_int = self._buf_ustr[o]
            # Now that we have the token we can read the variant value
            # Strings and signatures are the most common types
            # so we inline them for performance
            if token_as_int == TOKEN_O_AS_INT or token_as_int == TOKEN_S_AS_INT:
                headers[field_0] = self._read_string_unpack()
            elif token_as_int == TOKEN_G_AS_INT:
                headers[field_0] = self._read_signature()
            elif token_as_int == TOKEN_U_AS_INT:
                headers[field_0] = self._read_uint32_unpack()
            else:  # pragma: no cover
                token = self._buf_ustr[o : o + signature_len].decode()
                # There shouldn't be any other types in the header
                # but just in case, we'll read it using the slow path
                headers[field_0] = self._readers[token](
                    self, get_signature_tree(token).root_type
                )
        return headers

    def _read_header(self) -> None:
        """Read the header of the message."""
        # Signature is of the header is
        # BYTE, BYTE, BYTE, BYTE, UINT32, UINT32, ARRAY of STRUCT of (BYTE,VARIANT)
        self._read_to_pos(HEADER_SIGNATURE_SIZE)
        endian = self._buf_ustr[0]
        self._message_type = self._buf_ustr[1]
        self._flag = self._buf_ustr[2]
        protocol_version = self._buf_ustr[3]

        if protocol_version != PROTOCOL_VERSION:
            raise InvalidMessageError(
                f"got unknown protocol version: {protocol_version}"
            )

        if endian != LITTLE_ENDIAN and endian != BIG_ENDIAN:
            raise InvalidMessageError(
                f"Expecting endianness as the first byte, got {endian} from {self._buf}"
            )

        if cython.compiled:
            self._body_len = _ustr_uint32(self._buf_ustr, 4, endian)
            self._serial = _ustr_uint32(self._buf_ustr, 8, endian)
            self._header_len = _ustr_uint32(self._buf_ustr, 12, endian)
        elif endian == LITTLE_ENDIAN:
            (
                self._body_len,
                self._serial,
                self._header_len,
            ) = UNPACK_HEADER_LITTLE_ENDIAN(self._buf, 4)
            self._uint32_unpack = UINT32_UNPACK_LITTLE_ENDIAN
            self._int16_unpack = INT16_UNPACK_LITTLE_ENDIAN
            self._uint16_unpack = UINT16_UNPACK_LITTLE_ENDIAN
        else:  # BIG_ENDIAN
            self._body_len, self._serial, self._header_len = UNPACK_HEADER_BIG_ENDIAN(
                self._buf, 4
            )
            self._uint32_unpack = UINT32_UNPACK_BIG_ENDIAN
            self._int16_unpack = INT16_UNPACK_BIG_ENDIAN
            self._uint16_unpack = UINT16_UNPACK_BIG_ENDIAN

        if (
            self._body_len > _MAX_MESSAGE_SIZE
            or self._header_len > _MAX_MESSAGE_SIZE
            or self._body_len + self._header_len > _MAX_MESSAGE_SIZE
        ):
            raise InvalidMessageError(
                f"message size exceeds maximum {_MAX_MESSAGE_SIZE}: "
                f"header={self._header_len}, body={self._body_len}"
            )

        # align 8
        self._msg_len = self._header_len + (-self._header_len & 7) + self._body_len
        if self._endian != endian:
            self._readers = self._readers_by_type[endian]
            self._endian = endian

    def _read_body(self) -> None:
        """Read the body of the message."""
        self._read_to_pos(HEADER_SIGNATURE_SIZE + self._msg_len)
        self._pos = HEADER_ARRAY_OF_STRUCT_SIGNATURE_POSITION
        header_fields = self._header_fields(self._header_len)
        self._pos += -self._pos & 7  # align 8
        signature: str = header_fields[HEADER_SIGNATURE_IDX]
        if not self._body_len:
            tree = SIGNATURE_TREE_EMPTY
            body: list[Any] = []
        else:
            token_as_int = ord(signature[0])
            if len(signature) == 1:
                if token_as_int == TOKEN_O_AS_INT:
                    tree = SIGNATURE_TREE_O
                    body = [self._read_string_unpack()]
                elif token_as_int == TOKEN_S_AS_INT:
                    tree = SIGNATURE_TREE_S
                    body = [self._read_string_unpack()]
                else:
                    tree = get_signature_tree(signature)
                    body = [self._readers[t.token](self, t) for t in tree.types]
            elif token_as_int == TOKEN_S_AS_INT and signature == "sa{sv}as":
                tree = SIGNATURE_TREE_SA_SV_AS
                body = [
                    self._read_string_unpack(),
                    self.read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_1),
                    self.read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_2),
                ]
            elif token_as_int == TOKEN_O_AS_INT and signature == "oa{sa{sv}}":
                tree = SIGNATURE_TREE_OA_SA_SV
                body = [
                    self._read_string_unpack(),
                    self.read_array(SIGNATURE_TREE_OA_SA_SV_TYPES_1),
                ]
            elif token_as_int == TOKEN_O_AS_INT and signature == "oas":
                tree = SIGNATURE_TREE_OAS
                body = [
                    self._read_string_unpack(),
                    self.read_array(SIGNATURE_TREE_OAS_TYPES_1),
                ]
            elif token_as_int == TOKEN_A_AS_INT and signature == "a{oa{sa{sv}}}":
                tree = SIGNATURE_TREE_A_OA_SA_SV
                body = [self.read_array(SIGNATURE_TREE_A_OA_SA_SV_TYPES_0)]
            else:
                tree = get_signature_tree(signature)
                body = [self._readers[t.token](self, t) for t in tree.types]

        flags = MESSAGE_FLAG_MAP.get(self._flag)
        if flags is None:
            flags = MESSAGE_FLAG_INTENUM(self._flag)
        reply_serial = header_fields[HEADER_REPLY_SERIAL_IDX]
        message = Message.__new__(Message)
        message._fast_init(
            header_fields[HEADER_DESTINATION_IDX],
            header_fields[HEADER_PATH_IDX],
            header_fields[HEADER_INTERFACE_IDX],
            header_fields[HEADER_MEMBER_IDX],
            MESSAGE_TYPE_MAP[self._message_type],
            flags,
            header_fields[HEADER_ERROR_NAME_IDX],
            reply_serial if reply_serial is not None else 0,
            header_fields[HEADER_SENDER_IDX],
            self._unix_fds,
            tree,
            body,
            self._serial,
            # The D-Bus implementation already validates the message,
            # so we don't need to do it again.
            False,
        )
        self._message = message
        self._read_complete = True

    def unmarshall(self) -> Message | None:
        """Unmarshall the message.

        The underlying read function will raise BlockingIOError if the
        if there are not enough bytes in the buffer. This allows unmarshall
        to be resumed when more data comes in over the wire.
        """
        return self._unmarshall()

    def _unmarshall(self) -> Message | None:
        """Unmarshall the message.

        The underlying read function will raise BlockingIOError if the
        if there are not enough bytes in the buffer. This allows unmarshall
        to be resumed when more data comes in over the wire.
        """
        if self._read_complete:
            self._next_message()
        try:
            if not self._msg_len:
                self._read_header()
            self._read_body()
        except MARSHALL_STREAM_END_ERROR:
            return None
        return self._message

    _complex_parsers_unpack: dict[str, Callable[[Unmarshaller, SignatureType], Any]] = {
        "b": read_boolean,
        "o": read_string_unpack,
        "s": read_string_unpack,
        "g": read_signature,
        "a": read_array,
        "(": read_struct,
        "{": read_dict_entry,
        "v": read_variant,
        "h": read_uint32_unpack,
        UINT32_DBUS_TYPE: read_uint32_unpack,
        INT16_DBUS_TYPE: read_int16_unpack,
        UINT16_DBUS_TYPE: read_uint16_unpack,
    }

    _ctype_by_endian: dict[int, dict[str, READER_TYPE]] = {
        endian: build_simple_parsers(endian) for endian in (LITTLE_ENDIAN, BIG_ENDIAN)
    }

    _readers_by_type: dict[int, dict[str, READER_TYPE]] = {
        LITTLE_ENDIAN: {
            **_ctype_by_endian[LITTLE_ENDIAN],
            **_complex_parsers_unpack,
        },
        BIG_ENDIAN: {
            **_ctype_by_endian[BIG_ENDIAN],
            **_complex_parsers_unpack,
        },
    }



import asyncio
from unittest.mock import patch

import pytest

from dbus_fast import Message
from dbus_fast.aio import MessageBus


@pytest.mark.asyncio
async def test_bus_disconnect_before_reply():
    """In this test, the bus disconnects before the reply comes in. Make sure
    the caller receives a reply with the error instead of hanging."""
    bus = MessageBus()
    assert not bus.connected
    await bus.connect()
    assert bus.connected

    with patch.object(bus._writer, "_write_without_remove_writer"):
        ping = bus.call(
            Message(
                destination="org.freedesktop.DBus",
                path="/org/freedesktop/DBus",
                interface="org.freedesktop.DBus",
                member="Ping",
            )
        )

        asyncio.get_running_loop().call_soon(bus.disconnect)

        with pytest.raises((EOFError, BrokenPipeError)):
            await ping

    assert bus._disconnected
    assert not bus.connected

    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_unexpected_disconnect():
    bus = MessageBus()

    class FakeSocket:
        def send(self, *args, **kwargs):
            raise OSError

    assert not bus.connected
    await bus.connect()
    assert bus.connected

    with (
        patch.object(bus._writer, "_write_without_remove_writer"),
        patch.object(bus._writer, "sock", FakeSocket()),
    ):
        ping = bus.call(
            Message(
                destination="org.freedesktop.DBus",
                path="/org/freedesktop/DBus",
                interface="org.freedesktop.DBus",
                member="Ping",
            )
        )

        with pytest.raises(OSError):
            await ping

        assert bus._disconnected
        assert not bus.connected

    with pytest.raises(OSError):
        await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)

    bus.disconnect()
    with pytest.raises(OSError):
        await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)



from enum import Enum, IntFlag
from functools import cached_property


class BusType(Enum):
    """An enum that indicates a type of bus. On most systems, there are
    normally two different kinds of buses running.
    """

    SESSION = 1  #: A bus for the current graphical user session.
    SYSTEM = 2  #: A persistent bus for the whole machine.


class MessageType(Enum):
    """An enum that indicates a type of message."""

    METHOD_CALL = 1  #: An outgoing method call.
    METHOD_RETURN = 2  #: A return to a previously sent method call
    ERROR = 3  #: A return to a method call that has failed
    SIGNAL = 4  #: A broadcast signal to subscribed connections

    @cached_property
    def value(self) -> int:
        """Return the value."""
        return self._value_


MESSAGE_TYPE_MAP = {field.value: field for field in MessageType}


class MessageFlag(IntFlag):
    """Flags that affect the behavior of sent and received messages"""

    NONE = 0
    NO_REPLY_EXPECTED = 1  #: The method call does not expect a method return.
    NO_AUTOSTART = 2
    ALLOW_INTERACTIVE_AUTHORIZATION = 4

    @cached_property
    def value(self) -> int:
        """Return the value."""
        return self._value_


# This is written out because of https://github.com/python/cpython/issues/98976
MESSAGE_FLAG_MAP = {
    0: MessageFlag.NONE,
    1: MessageFlag.NO_REPLY_EXPECTED,
    2: MessageFlag.NO_AUTOSTART,
    4: MessageFlag.ALLOW_INTERACTIVE_AUTHORIZATION,
}


class NameFlag(IntFlag):
    """A flag that affects the behavior of a name request."""

    NONE = 0
    ALLOW_REPLACEMENT = 1  #: If another client requests this name, let them have it.
    REPLACE_EXISTING = 2  #: If another client owns this name, try to take it.
    DO_NOT_QUEUE = 4  #: Name requests normally queue and wait for the owner to release the name. Do not enter this queue.


class RequestNameReply(Enum):
    """An enum that describes the result of a name request."""

    PRIMARY_OWNER = 1  #: The bus owns the name.
    IN_QUEUE = 2  #: The bus is in a queue and may receive the name after it is relased by the primary owner.
    EXISTS = 3  #: The name has an owner and NameFlag.DO_NOT_QUEUE was given.
    ALREADY_OWNER = 4  #: The bus already owns the name.


class ReleaseNameReply(Enum):
    """An enum that describes the result of a name release request"""

    RELEASED = 1
    NON_EXISTENT = 2
    NOT_OWNER = 3


class PropertyAccess(Enum):
    """An enum that describes whether a DBus property can be gotten or set with
    the ``org.freedesktop.DBus.Properties`` interface.
    """

    READ = "read"  #: The property is readonly.
    WRITE = "write"  #: The property is writeonly.
    READWRITE = "readwrite"  #: The property can be read or written to.

    def readable(self) -> bool:
        """Get whether the property can be read."""
        return self == PropertyAccess.READ or self == PropertyAccess.READWRITE

    def writable(self) -> bool:
        """Get whether the property can be written to."""
        return self == PropertyAccess.WRITE or self == PropertyAccess.READWRITE


class ArgDirection(Enum):
    """For an introspected argument, indicates whether it is an input parameter or a return value."""

    IN = "in"
    OUT = "out"


class ErrorType(str, Enum):
    """An enum for the type of an error for a message reply.

    :seealso: http://man7.org/linux/man-pages/man3/sd-bus-errors.3.html
    """

    SERVICE_ERROR = "com.dubstepdish.dbus.next.ServiceError"  #: A custom error to indicate an exported service threw an exception.
    INTERNAL_ERROR = "com.dubstepdish.dbus.next.InternalError"  #: A custom error to indicate something went wrong with the library.
    CLIENT_ERROR = "com.dubstepdish.dbus.next.ClientError"  #: A custom error to indicate something went wrong with the client.

    FAILED = "org.freedesktop.DBus.Error.Failed"
    NO_MEMORY = "org.freedesktop.DBus.Error.NoMemory"
    SERVICE_UNKNOWN = "org.freedesktop.DBus.Error.ServiceUnknown"
    NAME_HAS_NO_OWNER = "org.freedesktop.DBus.Error.NameHasNoOwner"
    NO_REPLY = "org.freedesktop.DBus.Error.NoReply"
    IO_ERROR = "org.freedesktop.DBus.Error.IOError"
    BAD_ADDRESS = "org.freedesktop.DBus.Error.BadAddress"
    NOT_SUPPORTED = "org.freedesktop.DBus.Error.NotSupported"
    LIMITS_EXCEEDED = "org.freedesktop.DBus.Error.LimitsExceeded"
    ACCESS_DENIED = "org.freedesktop.DBus.Error.AccessDenied"
    AUTH_FAILED = "org.freedesktop.DBus.Error.AuthFailed"
    NO_SERVER = "org.freedesktop.DBus.Error.NoServer"
    TIMEOUT = "org.freedesktop.DBus.Error.Timeout"
    NO_NETWORK = "org.freedesktop.DBus.Error.NoNetwork"
    ADDRESS_IN_USE = "org.freedesktop.DBus.Error.AddressInUse"
    DISCONNECTED = "org.freedesktop.DBus.Error.Disconnected"
    INVALID_ARGS = "org.freedesktop.DBus.Error.InvalidArgs"
    FILE_NOT_FOUND = "org.freedesktop.DBus.Error.FileNotFound"
    FILE_EXISTS = "org.freedesktop.DBus.Error.FileExists"
    UNKNOWN_METHOD = "org.freedesktop.DBus.Error.UnknownMethod"
    UNKNOWN_OBJECT = "org.freedesktop.DBus.Error.UnknownObject"
    UNKNOWN_INTERFACE = "org.freedesktop.DBus.Error.UnknownInterface"
    UNKNOWN_PROPERTY = "org.freedesktop.DBus.Error.UnknownProperty"
    PROPERTY_READ_ONLY = "org.freedesktop.DBus.Error.PropertyReadOnly"
    UNIX_PROCESS_ID_UNKNOWN = "org.freedesktop.DBus.Error.UnixProcessIdUnknown"
    INVALID_SIGNATURE = "org.freedesktop.DBus.Error.InvalidSignature"
    INCONSISTENT_MESSAGE = "org.freedesktop.DBus.Error.InconsistentMessage"
    TIMED_OUT = "org.freedesktop.DBus.Error.TimedOut"
    MATCH_RULE_NOT_FOUND = "org.freedesktop.DBus.Error.MatchRuleNotFound"
    MATCH_RULE_INVALID = "org.freedesktop.DBus.Error.MatchRuleInvalid"
    INTERACTIVE_AUTHORIZATION_REQUIRED = (
        "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired"
    )
    INVALID_FILE_CONTENT = "org.freedesktop.DBus.Error.InvalidFileContent"
    SELINUX_SECURITY_CONTEXT_UNKNOWN = (
        "org.freedesktop.DBus.Error.SELinuxSecurityContextUnknown"
    )
    OBJECT_PATH_IN_USE = "org.freedesktop.DBus.Error.ObjectPathInUse"



from pytest_codspeed import BenchmarkFixture

from dbus_fast import Message

ITERATIONS = 1000

message = Message(
    destination="org.bluez",
    path="/",
    interface="org.freedesktop.DBus.ObjectManager",
    member="GetManagedObjects",
)


def test_marshall_bluez_get_managed_objects_message(
    benchmark: BenchmarkFixture,
) -> None:
    _marshall = message._marshall

    @benchmark
    def _():
        for _ in range(ITERATIONS):
            _marshall(False)



import os
from pathlib import Path
from unittest.mock import patch

import pytest

from dbus_fast._private.address import (
    get_bus_address,
    get_session_bus_address,
    get_system_bus_address,
    parse_address,
)
from dbus_fast.constants import BusType
from dbus_fast.errors import InvalidAddressError


def test_valid_addresses() -> None:
    valid_addresses = {
        "unix:path=/run/user/1000/bus": [("unix", {"path": "/run/user/1000/bus"})],
        "unix:abstract=/tmp/dbus-ft9sODWpZk,guid=a7b1d5912379c2d471165e9b5cb74a03": [
            (
                "unix",
                {
                    "abstract": "/tmp/dbus-ft9sODWpZk",
                    "guid": "a7b1d5912379c2d471165e9b5cb74a03",
                },
            )
        ],
        "unix1:key1=val1;unix2:key2=val2": [
            ("unix1", {"key1": "val1"}),
            ("unix2", {"key2": "val2"}),
        ],
        "unix:escaped=hello%20world": [("unix", {"escaped": "hello world"})],
        "tcp:host=127.0.0.1,port=55556": [
            ("tcp", {"host": "127.0.0.1", "port": "55556"})
        ],
        "unix:tmpdir=/tmp,;": [("unix", {"tmpdir": "/tmp"})],
    }

    for address, parsed in valid_addresses.items():
        assert parse_address(address) == parsed


def test_invalid_addresses() -> None:
    with pytest.raises(InvalidAddressError):
        assert parse_address("")
    with pytest.raises(InvalidAddressError):
        assert parse_address("unix")
    with pytest.raises(InvalidAddressError):
        assert parse_address("unix:tmpdir")
    with pytest.raises(InvalidAddressError):
        assert parse_address("unix:tmpdir=😁")


def test_get_system_bus_address() -> None:
    with patch.dict(os.environ, DBUS_SYSTEM_BUS_ADDRESS="unix:path=/dog"):
        assert get_system_bus_address() == "unix:path=/dog"
        assert get_bus_address(BusType.SYSTEM) == "unix:path=/dog"
    with patch.dict(os.environ, DBUS_SYSTEM_BUS_ADDRESS=""):
        assert get_system_bus_address() == "unix:path=/var/run/dbus/system_bus_socket"


def test_get_session_bus_address() -> None:
    with patch.dict(os.environ, DBUS_SESSION_BUS_ADDRESS="unix:path=/dog"):
        assert get_session_bus_address() == "unix:path=/dog"
        assert get_bus_address(BusType.SESSION) == "unix:path=/dog"
    with (
        patch.dict(os.environ, DBUS_SESSION_BUS_ADDRESS="", DISPLAY=""),
        pytest.raises(InvalidAddressError),
    ):
        assert get_session_bus_address()


def test_invalid_bus_address() -> None:
    with pytest.raises(Exception):
        assert get_bus_address(-1)


def test_session_bus_address_missing_display_raises() -> None:
    """Without DBUS_SESSION_BUS_ADDRESS and without DISPLAY, raise."""
    env = {
        k: v
        for k, v in os.environ.items()
        if k not in ("DBUS_SESSION_BUS_ADDRESS", "DISPLAY")
    }
    with (
        patch.dict(os.environ, env, clear=True),
        pytest.raises(InvalidAddressError, match=r"DISPLAY"),
    ):
        get_session_bus_address()


def test_session_bus_address_unparseable_display_raises() -> None:
    """A DISPLAY value that doesn't match the regex must raise InvalidAddressError."""
    with (
        patch.dict(os.environ, DBUS_SESSION_BUS_ADDRESS="", DISPLAY="not-a-display"),
        pytest.raises(InvalidAddressError, match=r"DISPLAY"),
    ):
        get_session_bus_address()


def _read_real_machine_id() -> str:
    """Return the host's real ``/var/lib/dbus/machine-id`` or skip the test."""
    try:
        with open("/var/lib/dbus/machine-id") as f:
            return f.read().rstrip()
    except OSError:
        pytest.skip("/var/lib/dbus/machine-id is not available on this host")


def _write_info_file(
    home: Path, machine_id: str, display_num: str, content: str
) -> Path:
    info_path = home / ".dbus" / "session-bus" / f"{machine_id}-{display_num}"
    info_path.parent.mkdir(parents=True, exist_ok=True)
    info_path.write_text(content)
    return info_path


def test_session_bus_address_from_dbus_info_file(tmp_path: Path) -> None:
    """Happy path: DISPLAY set, real machine-id, info file under HOME readable."""
    machine_id = _read_real_machine_id()
    expected_addr = "unix:abstract=/tmp/dbus-XYZ"
    _write_info_file(
        tmp_path,
        machine_id,
        "0",
        f"# comment\nDBUS_SESSION_BUS_ADDRESS={expected_addr}\nOTHER=foo\n",
    )
    with patch.dict(
        os.environ,
        DBUS_SESSION_BUS_ADDRESS="",
        DISPLAY=":0",
        HOME=str(tmp_path),
    ):
        assert get_session_bus_address() == expected_addr


def test_session_bus_address_strips_quotes_around_address(tmp_path: Path) -> None:
    """Quoted DBUS_SESSION_BUS_ADDRESS values are unquoted."""
    machine_id = _read_real_machine_id()
    _write_info_file(
        tmp_path,
        machine_id,
        "0",
        'DBUS_SESSION_BUS_ADDRESS="unix:path=/tmp/socket"\n',
    )
    with patch.dict(
        os.environ,
        DBUS_SESSION_BUS_ADDRESS="",
        DISPLAY="hostname:0.0",
        HOME=str(tmp_path),
    ):
        assert get_session_bus_address() == "unix:path=/tmp/socket"


def test_session_bus_address_info_file_missing_raises(tmp_path: Path) -> None:
    """If the dbus info file does not exist, raise InvalidAddressError."""
    _read_real_machine_id()
    with (
        patch.dict(
            os.environ,
            DBUS_SESSION_BUS_ADDRESS="",
            DISPLAY=":0",
            HOME=str(tmp_path),
        ),
        pytest.raises(InvalidAddressError, match=r"could not open dbus info file"),
    ):
        get_session_bus_address()


def test_session_bus_address_info_file_empty_value_raises(tmp_path: Path) -> None:
    """An empty DBUS_SESSION_BUS_ADDRESS= line in the info file must raise."""
    machine_id = _read_real_machine_id()
    _write_info_file(tmp_path, machine_id, "0", "DBUS_SESSION_BUS_ADDRESS=\n")
    with (
        patch.dict(
            os.environ,
            DBUS_SESSION_BUS_ADDRESS="",
            DISPLAY=":0",
            HOME=str(tmp_path),
        ),
        pytest.raises(InvalidAddressError, match=r"not set correctly"),
    ):
        get_session_bus_address()


def test_session_bus_address_info_file_missing_address_line_raises(
    tmp_path: Path,
) -> None:
    """An info file without a DBUS_SESSION_BUS_ADDRESS line must raise."""
    machine_id = _read_real_machine_id()
    _write_info_file(tmp_path, machine_id, "0", "OTHER=value\n# nothing useful\n")
    with (
        patch.dict(
            os.environ,
            DBUS_SESSION_BUS_ADDRESS="",
            DISPLAY=":0",
            HOME=str(tmp_path),
        ),
        pytest.raises(InvalidAddressError, match=r"could not find"),
    ):
        get_session_bus_address()



import os
import xml.etree.ElementTree as ET

import pytest

from dbus_fast import (
    ArgDirection,
    InvalidIntrospectionError,
    InvalidMemberNameError,
    PropertyAccess,
    SignatureType,
)
from dbus_fast import introspection as intr

with open(f"{os.path.dirname(__file__)}/data/strict-introspection.xml") as f:
    strict_data = f.read()

with open(f"{os.path.dirname(__file__)}/data/sloppy-introspection.xml") as f:
    sloppy_data = f.read()


def test_introspection_from_xml_sloppy():
    intr.Node.parse(sloppy_data, validate_property_names=False)


def test_introspection_from_xml_strict():
    try:
        node = intr.Node.parse(sloppy_data)
    except InvalidMemberNameError:
        pass
    else:
        assert False, "Expected an AssertionError"

    node = intr.Node.parse(strict_data)

    assert len(node.interfaces) == 1
    interface = node.interfaces[0]

    assert len(node.nodes) == 2
    assert len(interface.methods) == 3
    assert len(interface.signals) == 2
    assert len(interface.properties) == 1

    assert type(node.nodes[0]) is intr.Node
    assert node.nodes[0].name == "child_of_sample_object"
    assert type(node.nodes[1]) is intr.Node
    assert node.nodes[1].name == "another_child_of_sample_object"

    assert interface.name == "com.example.SampleInterface0"

    frobate = interface.methods[0]
    assert type(frobate) is intr.Method
    assert frobate.name == "Frobate"
    assert len(frobate.in_args) == 1
    assert len(frobate.out_args) == 2

    foo = frobate.in_args[0]
    assert type(foo) is intr.Arg
    assert foo.name == "foo"
    assert foo.direction == ArgDirection.IN
    assert foo.signature == "i"
    assert type(foo.type) is SignatureType
    assert foo.type.token == "i"

    bar = frobate.out_args[0]
    assert type(bar) is intr.Arg
    assert bar.name == "bar"
    assert bar.direction == ArgDirection.OUT
    assert bar.signature == "s"
    assert type(bar.type) is SignatureType
    assert bar.type.token == "s"

    prop = interface.properties[0]
    assert type(prop) is intr.Property
    assert prop.name == "Bar"
    assert prop.signature == "y"
    assert type(prop.type) is SignatureType
    assert prop.type.token == "y"
    assert prop.access == PropertyAccess.WRITE

    changed = interface.signals[0]
    assert type(changed) is intr.Signal
    assert changed.name == "Changed"
    assert len(changed.args) == 1
    new_value = changed.args[0]
    assert type(new_value) is intr.Arg
    assert new_value.name == "0-new_value"
    assert new_value.signature == "b"


def test_example_introspection_to_xml():
    node = intr.Node.parse(strict_data)
    tree = node.to_xml()
    assert tree.tag == "node"
    assert tree.attrib.get("name") == "/com/example/sample_object0"
    assert len(tree) == 3
    interface = tree[0]
    assert interface.tag == "interface"
    assert interface.get("name") == "com.example.SampleInterface0"
    assert len(interface) == 6
    method = interface[0]
    assert method.tag == "method"
    assert method.get("name") == "Frobate"
    assert len(method) == 4

    arg = method[0]
    assert arg.tag == "arg"
    assert arg.attrib.get("name") == "foo"
    assert arg.attrib.get("type") == "i"
    assert arg.attrib.get("direction") == "in"

    annotation = method[3]
    assert annotation.tag == "annotation"
    assert annotation.attrib.get("name") == "org.freedesktop.DBus.Deprecated"
    assert annotation.attrib.get("value") == "true"

    signal = interface[3]
    assert signal.tag == "signal"
    assert signal.attrib.get("name") == "Changed"
    assert len(signal) == 1

    arg = signal[0]
    assert arg.tag == "arg"
    assert arg.attrib.get("name") == "0-new_value"
    assert arg.attrib.get("type") == "b"

    signal = interface[4]
    assert signal.tag == "signal"
    assert signal.attrib.get("name") == "ChangedMulti"
    assert len(signal) == 2

    arg = signal[0]
    assert arg.tag == "arg"
    assert arg.attrib.get("name") == "new_value1"
    assert arg.attrib.get("type") == "b"

    arg = signal[1]
    assert arg.tag == "arg"
    assert arg.attrib.get("name") == "0-new_value2"
    assert arg.attrib.get("type") == "y"

    prop = interface[5]
    assert prop.attrib.get("name") == "Bar"
    assert prop.attrib.get("type") == "y"
    assert prop.attrib.get("access") == "write"


def test_default_interfaces():
    # just make sure it doesn't throw
    default = intr.Node.default()
    assert type(default) is intr.Node


def test_introspection_rejects_billion_laughs() -> None:
    """Nested-entity expansion in an Introspect reply is rejected."""
    payload = (
        '<?xml version="1.0"?>'
        "<!DOCTYPE node ["
        '<!ENTITY a "AAAAAAAAAA">'
        '<!ENTITY b "&a;&a;&a;&a;&a;&a;&a;&a;&a;&a;">'
        '<!ENTITY c "&b;&b;&b;&b;&b;&b;&b;&b;&b;&b;">'
        "]>"
        "<node>&c;</node>"
    )
    with pytest.raises(InvalidIntrospectionError, match="internal DTD"):
        intr.Node.parse(payload)


def test_introspection_rejects_quadratic_blowup() -> None:
    """A single large entity referenced many times is rejected."""
    payload = (
        '<?xml version="1.0"?>'
        "<!DOCTYPE node ["
        '<!ENTITY pad "' + ("A" * 1024) + '">'
        "]>"
        "<node>" + ("&pad;" * 64) + "</node>"
    )
    with pytest.raises(InvalidIntrospectionError, match="internal DTD"):
        intr.Node.parse(payload)


def test_introspection_rejects_xxe_external_entity() -> None:
    """External-entity declarations are rejected even without expansion."""
    payload = (
        '<?xml version="1.0"?>'
        "<!DOCTYPE node ["
        '<!ENTITY xxe SYSTEM "file:///etc/passwd">'
        "]>"
        "<node>&xxe;</node>"
    )
    with pytest.raises(InvalidIntrospectionError, match="internal DTD"):
        intr.Node.parse(payload)


def test_introspection_rejects_attlist_amplification() -> None:
    """ATTLIST default-value declarations are rejected (amplification vector)."""
    payload = (
        '<?xml version="1.0"?>'
        "<!DOCTYPE node ["
        '<!ATTLIST node bogus CDATA "' + ("A" * 1024) + '">'
        "]>"
        "<node/>"
    )
    with pytest.raises(InvalidIntrospectionError, match="internal DTD"):
        intr.Node.parse(payload)


def test_introspection_accepts_standard_doctype() -> None:
    """The standard D-Bus introspection DOCTYPE (no internal subset) parses."""
    payload = (
        '<?xml version="1.0"?>\n'
        '<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"\n'
        '"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">\n'
        '<node><interface name="org.example.Iface"/></node>'
    )
    node = intr.Node.parse(payload)
    assert len(node.interfaces) == 1
    assert node.interfaces[0].name == "org.example.Iface"


def test_introspection_malformed_xml_raises_parse_error() -> None:
    """Malformed XML still raises ET.ParseError, not a silent failure."""
    with pytest.raises(ET.ParseError):
        intr.Node.parse("<node><unclosed")


def test_introspection_parse_error_preserves_code_and_position() -> None:
    """ET.ParseError still carries the expat code and position metadata."""
    with pytest.raises(ET.ParseError) as excinfo:
        intr.Node.parse("<bad")
    assert excinfo.value.code  # non-zero expat error code
    assert isinstance(excinfo.value.position, tuple)
    assert len(excinfo.value.position) == 2


def test_introspection_rejects_default_namespaced_root() -> None:
    """A namespaced <node> root is still rejected by the root-tag check."""
    payload = '<node xmlns="urn:example"><interface name="a"/></node>'
    with pytest.raises(InvalidIntrospectionError, match="node"):
        intr.Node.parse(payload)
